using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;

namespace ModelEx.DirectX
{
    public class DirectXView : UserControl
    {
        protected   Color               m_xBackGroundColour = Color.Blue;
        protected   Device              m_xDevice;
        protected   PresentParameters   m_xPresentParams;
        private     int                 m_iLastDeviceReset;
        private     int                 m_iLastTick;
        private     int                 m_iLastFrameTime;
        private     int                 m_iFrameRate;
        private     int                 m_iLastFrameRate;
        private     bool                m_bPauseRendering;
        private     bool                m_bDeviceLost;

        public bool DeviceLost
        {
            get { return m_bDeviceLost; }
        }

        public int LastDeviceReset
        {
            get { return m_iLastDeviceReset; }
        }

        public bool PauseRendering
        {
            get { return m_bPauseRendering; }
            set
            {
                m_bPauseRendering = value;
                if (TopLevelControl != null &&
                    TopLevelControl.Focused == false)
                {
                    TopLevelControl.Focus();
                }
            }
        }

        public int FrameRate
        {
            get { return m_iLastFrameRate; }
        }

        public int LastFrameTime
        {
            get { return m_iLastFrameTime; }
        }

        public DirectXView()
        {
            SetStyle(ControlStyles.AllPaintingInWmPaint | ControlStyles.Opaque, true);
            Paint += new PaintEventHandler(Render);
        }

        protected override void OnGotFocus(EventArgs e)
        {
            base.OnGotFocus(e);
            Invalidate();
        }

        public virtual bool Initialise()
        {
            return SetupDevice();
        }

        protected virtual bool SetupDevice()
        {
            try
            {
                m_xPresentParams = new PresentParameters();
                m_xPresentParams.Windowed = true;
                m_xPresentParams.SwapEffect = SwapEffect.Discard;
                m_xPresentParams.EnableAutoDepthStencil = true;
                m_xPresentParams.AutoDepthStencilFormat = DepthFormat.D16;
                m_xPresentParams.DeviceWindow = this;

                m_xDevice = new Device(
                    0,
                    DeviceType.Hardware,
                    this.TopLevelControl,
                    CreateFlags.SoftwareVertexProcessing |
                    CreateFlags.MultiThreaded,
                    m_xPresentParams
                );

                m_xDevice.DeviceResizing += new CancelEventHandler(m_xDevice_DeviceResizing);
                m_xDevice.DeviceReset += new EventHandler(OnResetDevice);
                m_xDevice.DeviceLost += new EventHandler(OnDeviceLost);
            }
            catch (DirectXException)
            {
                return false;
            }

            return true;
        }

        void m_xDevice_DeviceResizing(object sender, CancelEventArgs e)
        {
            // e.Cancel = true;
        }

        protected virtual void OnDeviceLost(object sender, EventArgs e)
        {
            m_bDeviceLost = true;
        }

        protected virtual void OnResetDevice(object sender, EventArgs e)
        {
            m_bDeviceLost = false;
            m_iLastDeviceReset = Environment.TickCount;
        }

        protected void UpdateFrameRate()
        {
            // Track all frames in the last second...
            if (Environment.TickCount - m_iLastTick >= 1000)
            {
                m_iLastFrameRate = m_iFrameRate;
                m_iFrameRate = 0;
                m_iLastTick = Environment.TickCount;
                OnTick();
            }
            m_iLastFrameTime = Environment.TickCount;
            m_iFrameRate++;
        }

        protected virtual void OnTick()
        {
        }

        private void Render(object sender, PaintEventArgs e)
        {
            if (m_xDevice == null || TopLevelControl == null)
            {
                return;
            }

            if (((Form)TopLevelControl).WindowState == FormWindowState.Minimized ||
                ((Form)TopLevelControl).ContainsFocus == false)
            {
                return;
            }

            if (m_bDeviceLost)
            {
                int iCanReset = 0;
                bool bCanReset = !m_xDevice.CheckCooperativeLevel(out iCanReset);

                if (!bCanReset || (ResultCode)iCanReset != ResultCode.Success)
                {
                    return;
                }

                m_xDevice.Reset(m_xPresentParams);
            }

            // If the time since the last frame is less than a 30th of a second,
            //// then don't render the new frame...
            //if (((uint)Environment.TickCount - (uint)m_iLastFrameTime) < ((uint)3))
            //{
            //    return;
            //}

            UpdateFrameRate();

            // This should cause the other controls to be drawn...
            // No idea why it's needed though...
            TopLevelControl.Update();

            m_xDevice.Clear(ClearFlags.Target | ClearFlags.ZBuffer, m_xBackGroundColour, 1.0f, 0);
            RenderScene();
            m_xDevice.Present();

            // Queue the next frame...
            if (PauseRendering == false)
            {
                Invalidate();
            }
        }

        protected virtual void RenderScene()
        {
        }
    }
}
