﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class Mesh : Project.Resource
    {
        internal Vector3Source Positions;
        internal Vector3Source Normals;
        internal ColorValueSource Colors;
        internal Vector2Source TextureCoords;
        internal Material[] Materials;
        internal Triangles[] Triangles;

        internal Vector3 Centre;
        internal float Radius;
        internal Vector3 Scale;

        Mesh(Project project, String name) :
            base(project, name)
        {
        }

        ~Mesh()
        {
        }

        public static Mesh Create(Project project, String name)
        {
            Mesh mesh = project.FindMesh(name);
            return (mesh != null) ? mesh : new Mesh(project, name);
        }

        public override void Load(String folderName)
        {
        }

        public void ComputeBoundingSphere()
        {
            Vector3 averagePosition = new Vector3();
            float rcpNum = 1.0f / (float)Positions.Count; // Do this here as divides are far more expensive than multiplies.
            for (int p = 0; p < Positions.Count; p++)
            {
                // Instead of adding everything up and then dividing by the number (which could lead  
                // to overflows) I'll divide by the number as I go along.  The result is the same. 
                averagePosition.X += Positions[p].X * rcpNum;
                averagePosition.Y += Positions[p].Y * rcpNum;
                averagePosition.Z += Positions[p].Z * rcpNum;
            }

            float maxDistanceSquared = 0.0f;

            for (int p = 0; p < Positions.Count; p++)
            {
                Vector3 offset = averagePosition - Positions[p];

                // Note we may as well use the square length as the sqrt is very expensive and the  
                // maximum square length will ALSO be the maximum length and yet we only need to 
                // do one sqrt this way :)
                float distanceSquared = offset.LengthSq();
                if (distanceSquared > maxDistanceSquared)
                {
                    maxDistanceSquared = distanceSquared;
                }
            }

            float radius = (float)Math.Sqrt(maxDistanceSquared);

            Centre = averagePosition;
            Radius = radius;
            float scale = 1.0f / Radius;
            Scale = new Vector3(scale, scale, scale);
        }

        public void Render(Matrix world, RenderFlags renderFlags, CullType cullType)
        {
            if (Triangles == null)
            {
                return;
            }

            Device device = Project.Device;

            for (int m = 0; m < Triangles.Length; m++)
            {
                D3DX.Effect effect = Triangles[m].Material.Effect.D3DXEffect;

                effect.Technique = effect.GetTechnique(1);

                #region RenderStates
                #region Lighting
                bool lighting = false;
                if ((renderFlags & RenderFlags.Lighting) != 0)
                {
                    lighting = (bool)Triangles[m].Material.GetParameter(6);
                }
                effect.SetValue("bLighting", lighting);
                #endregion

                #region Textures
                TextureOperation colorOperation = TextureOperation.SelectArg2;
                TextureOperation alphaOperation = TextureOperation.SelectArg2;
                if ((renderFlags & RenderFlags.Textures) != 0)
                {
                    colorOperation = (TextureOperation)Triangles[m].Material.GetParameter(7);
                    alphaOperation = TextureOperation.Modulate;
                }
                effect.SetValue("iColorOperation", (int)colorOperation);
                effect.SetValue("iAlphaOperation", (int)alphaOperation);
                #endregion

                #region Texture Filtering
                TextureFilter textureFilter = TextureFilter.Point;
                if ((renderFlags & RenderFlags.TextureFiltering) != 0)
                {
                    textureFilter = (TextureFilter)Triangles[m].Material.GetParameter(8);
                }
                effect.SetValue("iTextureFilter", (int)textureFilter);
                #endregion

                #region Wireframe Mode
                FillMode fillMode = (FillMode)Triangles[m].Material.GetParameter(9);
                if ((renderFlags & RenderFlags.WireFrame) != 0)
                {
                    fillMode = FillMode.WireFrame;
                }
                effect.SetValue("iFillMode", (int)fillMode);
                #endregion

                #region Cull Mode
                effect.SetValue("iCullMode", (int)cullType);
                #endregion

                #region Transforms
                effect.SetValue("xWorld", world);
                effect.SetValue("xView", device.Transform.View);
                effect.SetValue("xWorldViewProj", world * device.Transform.View * device.Transform.Projection);
                #endregion
                #endregion

                effect.ApplyParameterBlock(Triangles[m].Material.ParameterBlock);

                //EffectHandle handle = effect.GetTechnique(0);
                //if (handle == null || !effect.IsTechniqueValid(handle))
                //{
                //    throw new Exception("Technique is not valid");
                //}

                int passes = effect.Begin(FX.None);

                for (int p = 0; p < passes; p++)
                {
                    effect.BeginPass(p);

                    #region Rendering
                    device.VertexFormat = Triangles[m].VertexFormat;
                    if (Triangles[m].VertexBuffer != null && Triangles[m].IndexBuffer != null)
                    {
                        device.SetStreamSource(0, Triangles[m].VertexBuffer, 0, Triangles[m].Stride);
                        device.Indices = Triangles[m].IndexBuffer;
                        device.DrawIndexedPrimitives(PrimitiveType.TriangleList, 0, 0, Triangles[m].VertexCount, 0, Triangles[m].IndexCount / 3);
                        //device.DrawPrimitives(PrimitiveType.TriangleList, 0, Triangles[m].vertexCount / 3);
                    }
                    #endregion

                    effect.EndPass();
                }

                effect.End();
            }
        }
    }
}
