﻿using System;
using System.Collections.Generic;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    partial class Project
    {
        class ArraySet<T>
        {
            int ArraySize;
            int LastArraySize;
            public readonly int NumEntries;
            public readonly int NumArrays;
            public readonly T[][] Arrays;

            public ArraySet(int arraySize, int numEntries)
            {
                ArraySize = arraySize;
                NumEntries = numEntries;
                NumArrays = numEntries / ArraySize;
                LastArraySize = numEntries % ArraySize;
                if (LastArraySize != 0)
                {
                    NumArrays++;
                }
                else
                {
                    LastArraySize = ArraySize;
                }

                Arrays = new T[NumArrays][];
                for (int a = 0; a < NumArrays; a++)
                {
                    if (a == NumArrays - 1)
                    {
                        Arrays[a] = new T[LastArraySize];
                    }
                    else
                    {
                        Arrays[a] = new T[ArraySize];
                    }
                }
            }

            public T this[int i]
            {
                get { return Arrays[i / ArraySize][i % ArraySize]; }
                set { Arrays[i / ArraySize][i % ArraySize] = value; }
            }
        }

        public abstract class Resource
        {
            public Project Project;
            public readonly String Name;
            public readonly UInt32 HashID;
            internal Device Device { get { return Project.Device; } }

            protected Resource(Project project, String name)
            {
                Project = project;
                Name = name;
                HashID = Project.GetHashID(name);

                Type type = GetType();
                List<Resource> references = Project.GetResourceList(type);
                if (references != null)
                {
                    references.Add(this);
                }
            }

            ~Resource()
            {
            }

            public abstract void Load(String folderName);
        }

        internal Device Device;

        List<Resource> EffectReferences;
        List<Resource> MeshReferences;
        List<Resource> TextureReferences;
        List<Resource> MaterialReferences;
        List<Resource> VertexBufferReferences;
        List<Resource> IndexBufferReferences;
        List<Resource> PhysicalReferences;

        String _Name;
        public String Name
        {
            get { return (String)_Name.Clone(); }
        }

        Effect DefaultEffect;

        public Project(Device device)
        {
            Device = device;

            EffectReferences = new List<Resource>();
            MeshReferences = new List<Resource>();
            TextureReferences = new List<Resource>();
            MaterialReferences = new List<Resource>();
            VertexBufferReferences = new List<Resource>();
            IndexBufferReferences = new List<Resource>();
            PhysicalReferences = new List<Resource>();

            _Name = "New";

            DefaultEffect = Effect.Create(this, "default");
            DefaultEffect.LoadFromMemory("ModelEx_Default.fx");
            D3DX.Effect effect = DefaultEffect.D3DXEffect;

            EffectHandle lightsArray = effect.GetParameter(null, "xLights");
            EffectHandle lights_0 = effect.GetParameterElement(lightsArray, 0);

            EffectHandle lights_0_Position = effect.GetParameter(lights_0, "Position");
            EffectHandle lights_0_Direction = effect.GetParameter(lights_0, "Direction");
            EffectHandle lights_0_Diffuse = effect.GetParameter(lights_0, "Diffuse");
            EffectHandle lights_0_Specular = effect.GetParameter(lights_0, "Specular");
            EffectHandle lights_0_Ambient = effect.GetParameter(lights_0, "Ambient");
            EffectHandle lights_0_Attenuation = effect.GetParameter(lights_0, "Attenuation");

            effect.SetValue(lights_0_Position, new Vector4(0.0f, 0.0f, 0.0f, 0.0f));
            effect.SetValue(lights_0_Direction, new Vector4(0.0f, -0.3f, 0.5f, 1.0f));
            effect.SetValue(lights_0_Diffuse, new Vector4(1.0f, 1.0f, 1.0f, 1.0f));
            effect.SetValue(lights_0_Specular, new Vector4(0.0f, 0.0f, 0.0f, 0.0f));
            effect.SetValue(lights_0_Ambient, new Vector4(0.5f, 0.5f, 0.5f, 1.0f));
            effect.SetValue(lights_0_Attenuation, new Vector4(0.0f, 0.0f, 0.0f, 0.0f));
        }

        List<Resource> GetResourceList(Type type)
        {
            List<Resource> references = null;

            if (type == typeof(Effect))
            {
                references = EffectReferences;
            }
            else if (type == typeof(Mesh))
            {
                references = MeshReferences;
            }
            else if (type == typeof(Texture))
            {
                references = TextureReferences;
            }
            else if (type == typeof(Material))
            {
                references = MaterialReferences;
            }
            else if (type == typeof(Physical))
            {
                references = PhysicalReferences;
            }
            else if (type == typeof(Triangles))
            {
                references = IndexBufferReferences;
            }

            return references;
        }

        T FindResource<T>(String name) where T : Resource
        {
            UInt32 hashID = GetHashID(name);
            return FindResource<T>(hashID);
        }

        T FindResource<T>(UInt32 hashID) where T : Resource
        {
            List<Resource> references = GetResourceList(typeof(T));

            if (references == null)
            {
                return null;
            }

            foreach (Resource Reference in references)
            {
                T resource = (T)Reference;
                if (resource.HashID == hashID)
                {
                    return resource;
                }
            }

            return null;
        }

        public static UInt32 GetHashID(String input)
        {
            // Bob Jenkins one-at-a-time hash function
            // http://burtleburtle.net/bob/hash/doobs.html
            UInt32 output = 0;
            for (int i = 0; i < input.Length; i++)
            {
                output += input[i];
                output += (output << 10);
                output ^= (output >> 6);
            }

            output += (output << 3);
            output ^= (output >> 11);
            output += (output << 15);

            return output;
        }

        public Effect FindEffect(String name)
        {
            return FindResource<Effect>(name);
        }

        public String ListEffects()
        {
            String resources = "";

            List<Resource> references = GetResourceList(typeof(Effect));

            if (references == null)
            {
                return null;
            }

            foreach (Resource Reference in references)
            {
                Resource resource = (Resource)Reference;
                resources += resource.Name;
                resources += "\r\n";
            }

            return resources;
        }

        public Mesh FindMesh(String name)
        {
            return FindResource<Mesh>(name);
        }

        public Texture FindTexture(String name)
        {
            return FindResource<Texture>(name);
        }

        public String ListTextures()
        {
            String resources = "";

            List<Resource> references = GetResourceList(typeof(Texture));

            if (references == null)
            {
                return null;
            }

            foreach (Resource Reference in references)
            {
                Resource resource = (Resource)Reference;
                resources += resource.Name;
                resources += "\r\n";
            }

            return resources;
        }

        public Material FindMaterial(String name)
        {
            return FindResource<Material>(name);
        }

        public String ListMaterials()
        {
            String resources = "";

            List<Resource> references = GetResourceList(typeof(Material));

            if (references == null)
            {
                return null;
            }

            foreach (Resource Reference in references)
            {
                Material resource = (Material)Reference;
                resources += resource.Name;
                resources += "    ";
                resources += resource.Texture.Name;
                resources += "    ";
                resources += resource.Effect.Name;
                resources += "\r\n";
            }

            return resources;
        }

        public Triangles FindTriangles(String name)
        {
            return FindResource<Triangles>(name);
        }

        public Physical FindObject(String name)
        {
            return FindResource<Physical>(name);
        }

        public void LoadResources<T>(String folderName) where T : Resource
        {
            List<Resource> references = GetResourceList(typeof(T));

            if (references == null)
            {
                return;
            }

            foreach (Resource Reference in references)
            {
                T resource = (T)Reference;
                resource.Load(folderName);
            }
        }

        public void LoadEffects(String folderName)
        {
            LoadResources<Effect>(folderName);
        }

        public void LoadMeshes(String folderName)
        {
            LoadResources<Mesh>(folderName);
        }

        public void LoadTextures(String folderName)
        {
            LoadResources<Texture>(folderName);
        }

        public void RenderObjects(RenderFlags renderFlags, CullType cullType, float scale)
        {
            if (PhysicalReferences == null)
            {
                return;
            }

            foreach (Resource Reference in PhysicalReferences)
            {
                Physical currentObject = (Physical)Reference;
                currentObject.Render(renderFlags, cullType, scale);
            }
        }
    }
}
