﻿using System;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    abstract class Source
    {
        public readonly String Name;
        public readonly UInt32 HashID;
        public readonly Int32 Stride;
        public readonly Int32 Count;
        public readonly String[] ParameterNames;
        public readonly DeclarationUsage Usage;

        protected Source(String name, DeclarationUsage usage, Int32 stride, Int32 count)
        {
            Name = name;
            HashID = Project.GetHashID(name);
            Stride = stride;
            Count = count;
            ParameterNames = new String[Stride];
            Usage = usage;
        }

        ~Source()
        {
        }
    }

    abstract class Int32SourceBase : Source
    {
        public readonly Int32[] Ints;

        protected Int32SourceBase(String name, DeclarationUsage usage, Int32 stride, Int32 count) :
            base(name, usage, stride, count)
        {
            Ints = new Int32[Stride * Count];
        }
    }

    class Int32Source : Int32SourceBase
    {
        protected Int32Source(String name, DeclarationUsage usage, Int32 count, String x) :
            base(name, usage, 1, count)
        {
            ParameterNames[0] = x;
        }

        public static Int32Source Create(String name, DeclarationUsage usage, Int32 count, String x)
        {
            return new Int32Source(name, usage, count, x);
        }

        public Int32 this[Int32 i]
        {
            get
            {
                return Ints[i];
            }

            set
            {
                Ints[i] = value;
            }
        }
    }

    abstract class FloatSourceBase : Source
    {
        public readonly float[] Floats;

        protected FloatSourceBase(String name, DeclarationUsage usage, Int32 stride, Int32 count) :
            base(name, usage, stride, count)
        {
            Floats = new float[Stride * Count];
        }
    }

    class FloatSource : FloatSourceBase
    {
        protected FloatSource(String name, DeclarationUsage usage, Int32 count, String x) :
            base(name, usage, 1, count)
        {
            ParameterNames[0] = x;
        }

        public static FloatSource Create(String name, DeclarationUsage usage, Int32 count, String x)
        {
            return new FloatSource(name, usage, count, x);
        }

        public float this[Int32 i]
        {
            get
            {
                return Floats[i];
            }

            set
            {
                Floats[i] = value;
            }
        }
    }

    class Vector2Source : FloatSourceBase
    {
        protected Vector2Source(String name, DeclarationUsage usage, Int32 count, String x, String y) :
            base(name, usage, 2, count)
        {
            ParameterNames[0] = x;
            ParameterNames[1] = y;
        }

        public static Vector2Source Create(String name, DeclarationUsage usage, Int32 count, String x, String y)
        {
            return new Vector2Source(name, usage, count, x, y);
        }

        public Vector2 this[Int32 i]
        {
            get
            {
                i *= Stride;
                Vector2 value = new Vector2();
                value.X = Floats[i];
                value.Y = Floats[i + 1];
                return value;
            }

            set
            {
                i *= Stride;
                Floats[i] = value.X;
                Floats[i + 1] = value.Y;
            }
        }
    }

    class Vector3Source : FloatSourceBase
    {
        protected Vector3Source(String name, DeclarationUsage usage, Int32 count, String x, String y, String z) :
            base(name, usage, 3, count)
        {
            ParameterNames[0] = x;
            ParameterNames[1] = y;
            ParameterNames[2] = z;
        }

        public static Vector3Source Create(String name, DeclarationUsage usage, Int32 count, String x, String y, String z)
        {
            return new Vector3Source(name, usage, count, x, y, z);
        }

        public Vector3 this[Int32 i]
        {
            get
            {
                i *= Stride;
                Vector3 value = new Vector3();
                value.X = Floats[i];
                value.Y = Floats[i + 1];
                value.Z = Floats[i + 2];
                return value;
            }

            set
            {
                i *= Stride;
                Floats[i] = value.X;
                Floats[i + 1] = value.Y;
                Floats[i + 2] = value.Z;
            }
        }
    }

    class ColorValueSource : FloatSourceBase
    {
        protected ColorValueSource(String name, DeclarationUsage usage, Int32 count, String x, String y, String z) :
            base(name, usage, 3, count)
        {
            ParameterNames[0] = x;
            ParameterNames[1] = y;
            ParameterNames[2] = z;
        }

        public static ColorValueSource Create(String name, DeclarationUsage usage, Int32 count, String x, String y, String z)
        {
            return new ColorValueSource(name, usage, count, x, y, z);
        }

        public ColorValue this[Int32 i]
        {
            get
            {
                i *= Stride;
                ColorValue value = new ColorValue();
                value.Alpha = 1.0f;
                value.Red = Floats[i];
                value.Green = Floats[i + 1];
                value.Blue = Floats[i + 2];
                return value;
            }

            set
            {
                i *= Stride;
                Floats[i] = value.Red;
                Floats[i + 1] = value.Green;
                Floats[i + 2] = value.Blue;
            }
        }
    }
}