﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Data;
using System.IO;
using System.Text;
using System.Windows.Forms;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;
using Imaging = System.Drawing.Imaging;

namespace ModelEx
{
    class ModelExView : DirectX.DirectXView
    {
        #region Enumerations

        public enum ControlType
        {
            None = 0,
            OrbitCamera = 1,
            FlyCamera = 2,
            MapCamera = 3,
            RotateObject = 4,
        }

        public enum ModelFormat
        {
            Binary = D3DX.XFileFormat.Binary,
            Text = D3DX.XFileFormat.Text,
            BinaryCompressed = D3DX.XFileFormat.Binary | D3DX.XFileFormat.Compressed,
            TextCompressed = D3DX.XFileFormat.Text | D3DX.XFileFormat.Compressed,
            LastUsed = 4
        }

        public enum ImageFormat
        {
            Bmp = 0,
            Jpeg = 1,
            Png = 2,
            LastUsed = 3
        }

        #endregion

        #region Members

        RenderFlags RenderFlags;
        CullType CullMode;
        ControlType ControlMode;
        DebugType DebugMode;

        Project Project;
        Physical SelectedObject;
        Camera Camera;
        ArcBall ArcBall;
        Point LastMousePosition;
        Point LastMouseDownPosition;

        ModelExFont ArielFont;
        ModelExFont Wingdings2Font;

        ModelExListBox Menu;
        ModelExTextBox DebugOutput;
        ModelExTextBox TextureOutput;
        ModelExTextBox MaterialOutput;
        ModelExTextBox EffectOutput;

        ModelExScrollBar ScrollBar;

        #endregion

        public ModelExView() : base()
        {
            m_xBackGroundColour = Color.Gray;
            LastMousePosition = new Point();
            MouseUp += new MouseEventHandler(ModelExView_MouseUp);
            MouseDown += new MouseEventHandler(ModelExView_MouseDown);
            MouseMove += new MouseEventHandler(ModelExView_MouseMove);
            MouseWheel += new MouseEventHandler(ModelExView_MouseWheel);
            PreviewKeyDown += new PreviewKeyDownEventHandler(ModelExView_PreviewKeyDown);
            GotFocus += new EventHandler(ModelExView_ChangeFocus);
            LostFocus += new EventHandler(ModelExView_ChangeFocus);
        }

        #region Device States

        public override bool Initialise()
        {
            if (!base.Initialise())
            {
                return false;
            }

            SetupRenderStates();
            SetupLighting();
            SetupCamera();
            SetupControlStates();

            SetupFonts();
            SetupDebugOutput();
            SetupMenus();

            Project = new Project(m_xDevice);

            return true;
        }

        protected override void OnResetDevice(object sender, EventArgs e)
        {
            base.OnResetDevice(sender, e);
            SetupLighting();
            Update();
            UpdateControlStates();

            // May need to reload the mesh with MeshFlags.SystemMemory
            // May need to reload the textures
            // MessageBox.Show("Where is my device?!");
        }

        private void SetupRenderStates()
        {
            CullMode = CullType.Clockwise;
            Lighting = true;
            Textures = true;
            TextureFiltering = true;
            Wireframe = false;
            Scaling = true;
        }

        private void SetupLighting()
        {
            m_xDevice.RenderState.Lighting = true;
            m_xDevice.RenderState.Ambient = Color.FromArgb(unchecked((int)0xFF808080));
            m_xDevice.Lights[0].Type = LightType.Directional;
            m_xDevice.Lights[0].Diffuse = Color.FromArgb(unchecked((int)0xFFFFFFFF));
            m_xDevice.Lights[0].Direction = new Vector3(0.0f, -0.3f, 0.5f);
            m_xDevice.Lights[0].Direction.Normalize();
            m_xDevice.Lights[0].Enabled = false; // Can't set true if already true?
            m_xDevice.Lights[0].Enabled = true;
            //m_xDevice.Lights[0].Position = new Vector3(0.0f, 0.0f, -10000.0f);
        }

        private void SetupCamera()
        {
            Camera = new Camera(
                ref m_xDevice,
                new Vector3(0.0f, 0.0f, 0.0f),
                2.5f
            );

            UpdateControlStates();
        }

        private void SetupControlStates()
        {
            ControlMode = ControlType.OrbitCamera;
        }

        private void UpdateControlStates()
        {
            switch (ControlMode)
            {
                case ControlType.OrbitCamera:
                    if (Camera == null)
                    {
                        goto default;
                    }
                    Camera.Mode = Camera.CameraMode.Orbit;
                    Camera.Target = new Vector3(0.0f, 0.0f, 0.0f);
                    ArcBall = Camera.ArcBall;
                    break;
                case ControlType.FlyCamera:
                    if (Camera == null)
                    {
                        goto default;
                    }
                    Camera.Mode = Camera.CameraMode.Fly;
                    ArcBall = Camera.ArcBall;
                    break;
                case ControlType.MapCamera:
                    if (Camera == null)
                    {
                        goto default;
                    }
                    Camera.Mode = Camera.CameraMode.Map;
                    ArcBall = Camera.ArcBall;
                    break;
                case ControlType.RotateObject:
                    if (SelectedObject == null)
                    {
                        goto default;
                    }
                    Camera.Mode = Camera.CameraMode.Orbit;
                    Camera.Target = new Vector3(0.0f, 0.0f, 0.0f);
                    ArcBall = SelectedObject.ArcBall;
                    break;
                default:
                    Camera.Mode = Camera.CameraMode.Manual;
                    ArcBall = null;
                    break;
            }
        }

        private void SetupFonts()
        {
            ArielFont = new ModelExFont(m_xDevice, "Ariel", 9.5f);
            Wingdings2Font = new ModelExFont(m_xDevice, "Wingdings 2", 9.5f);
        }

        private void SetupMenus()
        {
            String[] statusText = { "", "\uF050" };

            ModelExListBox xTexturesMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 40), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xTexturesMenu.AddItem("From File...", 0, null, new ModelExListBox.ItemClickHandler(ModelExView_FromFile));
            xTexturesMenu.AddItem("From Folder...", 1, null, new ModelExListBox.ItemClickHandler(ModelExView_FromFolder));

            ModelExListBox xFileMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 60), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xFileMenu.AddItem("Open", 0, null, new ModelExListBox.ItemClickHandler(ModelExView_Open));
            xFileMenu.AddItem("Save", 1, null, new ModelExListBox.ItemClickHandler(ModelExView_Save));
            xFileMenu.AddItem("Load Textures", 2, xTexturesMenu);
            xFileMenu.AddItem("Exit", 3, null, new ModelExListBox.ItemClickHandler(ModelExView_Exit));

            ModelExListBox xCullMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 40), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xCullMenu.AddItem("Clockwise", 0, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_SelectCullMode));
            xCullMenu.AddItem("Anti-Clockwise", 1, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectCullMode));
            xCullMenu.AddItem("None", 2, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectCullMode));

            ModelExListBox xViewMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 60), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xViewMenu.AddItem("Reset All", 0, null, new ModelExListBox.ItemClickHandler(ModelExView_Exit));
            xViewMenu.AddItem("Reset Camera Position", 1, null, new ModelExListBox.ItemClickHandler(ResetCameraPosition));
            xViewMenu.AddItem("Reset Object Position", 2, null, new ModelExListBox.ItemClickHandler(ModelExView_Exit));
            xViewMenu.AddItem("Cull Mode", 3, xCullMenu);
            xViewMenu.AddItem("Lighting/Shading", 4, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_ToggleLighting));
            xViewMenu.AddItem("Show Textures", 5, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_ToggleTextures));
            xViewMenu.AddItem("Texture Filtering", 6, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_ToggleTextureFiltering));
            xViewMenu.AddItem("Wireframe", 7, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_ToggleWireframe));
            xViewMenu.AddItem("Scaling", 8, null, new ModelExListBox.ItemClickHandler(ModelExView_Exit));
            xViewMenu.AddItem("Model", 9, null, new ModelExListBox.ItemClickHandler(ModelExView_Exit));

            //disable entries that are not implemented
            xViewMenu.Items[0].Enabled = false;
            xViewMenu.Items[2].Enabled = false;
            xViewMenu.Items[8].Enabled = false;
            xViewMenu.Items[9].Enabled = false;

            ModelExListBox xDisplayMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 60), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xDisplayMenu.AddItem("Debug", 0, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectDebugOutput));
            xDisplayMenu.AddItem("Textures", 1, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectDebugOutput));
            xDisplayMenu.AddItem("Effects", 2, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectDebugOutput));
            xDisplayMenu.AddItem("Materials", 3, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectDebugOutput));
            xDisplayMenu.AddItem("Hide", 4, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_SelectDebugOutput));

            ModelExListBox xControlMenu = new ModelExListBox(m_xDevice, new Rectangle(0, 20, 150, 60), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 150, 20, true);
            xControlMenu.AddItem("Orbit Cam", 0, null, statusText, 1, new ModelExListBox.ItemClickHandler(ModelExView_SelectControlMode));
            xControlMenu.AddItem("Fly Cam", 1, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectControlMode));
            xControlMenu.AddItem("Map Cam", 2, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectControlMode));
            xControlMenu.AddItem("Rotate Object", 3, null, statusText, 0, new ModelExListBox.ItemClickHandler(ModelExView_SelectControlMode));

            Menu = new ModelExListBox(m_xDevice, new Rectangle(0, 0, 500, 20), 0xFF400000, 0xFF400000, 0xFFFFFFFF, 0xFF808080, ArielFont, Wingdings2Font, new Point(6, 2), 60, 20, false);
            Menu.AddItem("File", 0, xFileMenu);
            Menu.AddItem("View", 1, xViewMenu);
            Menu.AddItem("Display", 2, xDisplayMenu);
            Menu.AddItem("Control", 3, xControlMenu);
        }

        private void SetupDebugOutput()
        {
            DebugMode = DebugType.None;
            DebugOutput = new ModelExTextBox(m_xDevice, new Rectangle(50, 50, 500, 150), 0xFF000040, 0xFF000040, 0xFFFFFFFF, 0xFF808080, ArielFont, new Point(6, 6));
            TextureOutput = new ModelExTextBox(m_xDevice, new Rectangle(50, 50, 500, 150), 0xFF000040, 0xFF000040, 0xFFFFFFFF, 0xFF808080, ArielFont, new Point(6, 6));
            MaterialOutput = new ModelExTextBox(m_xDevice, new Rectangle(50, 50, 500, 150), 0xFF000040, 0xFF000040, 0xFFFFFFFF, 0xFF808080, ArielFont, new Point(6, 6));
            EffectOutput = new ModelExTextBox(m_xDevice, new Rectangle(50, 50, 500, 150), 0xFF000040, 0xFF000040, 0xFFFFFFFF, 0xFF808080, ArielFont, new Point(6, 6));

            ScrollBar = new ModelExScrollBar(m_xDevice, new Rectangle(0, 0, 15, 150), 0xFF000080, 0xFF0000FF, 0xFFFFFFFF, 0xFF000040, true, 2000);
        }

        #endregion

        #region View Modes

        public bool ResetRenderStates(ModelExListBox listBox)
        {
            SetupRenderStates();
            return true;
        }

        public bool ResetCameraPosition(ModelExListBox listBox)
        {
            SetupCamera();
            return true;
        }

        public bool ModelExView_SelectCullMode(ModelExListBox listBox)
        {
            for (int i = 0; i < listBox.Items.Count; i++ )
            {
                if (listBox.Items[i] == listBox.SelectedItem)
                {
                    listBox.Items[i].StatusBox.Status = 1;
                    switch (i)
                    {
                        case 0:
                            CullMode = CullType.Clockwise;
                            break;
                        case 1:
                            CullMode = CullType.CounterClockwise;
                            break;
                        case 2:
                            CullMode = CullType.None;
                            break;
                    }
                }
                else
                {
                    listBox.Items[i].StatusBox.Status = 0;
                }
            }

            return true;
        }

        public bool ModelExView_SelectControlMode(ModelExListBox listBox)
        {
            for (int i = 0; i < listBox.Items.Count; i++)
            {
                if (listBox.Items[i] == listBox.SelectedItem)
                {
                    listBox.Items[i].StatusBox.Status = 1;
                    switch (i)
                    {
                        case 0:
                            ControlMode = ControlType.OrbitCamera;
                            break;
                        case 1:
                            ControlMode = ControlType.FlyCamera;
                            break;
                        case 2:
                            ControlMode = ControlType.MapCamera;
                            break;
                        case 3:
                            ControlMode = ControlType.RotateObject;
                            break;
                    }
                }
                else
                {
                    listBox.Items[i].StatusBox.Status = 0;
                }
            }

            return true;
        }

        public bool Lighting
        {
            get { return ((RenderFlags & RenderFlags.Lighting) != 0); }
            set { if (Lighting != value) RenderFlags ^= RenderFlags.Lighting; }
        }

        public bool Textures
        {
            get { return ((RenderFlags & RenderFlags.Textures) != 0); }
            set { if (Textures != value) RenderFlags ^= RenderFlags.Textures; }
        }

        public bool TextureFiltering
        {
            get { return ((RenderFlags & RenderFlags.TextureFiltering) != 0); }
            set { if (TextureFiltering != value) RenderFlags ^= RenderFlags.TextureFiltering; }
        }

        public bool Wireframe
        {
            get { return ((RenderFlags & RenderFlags.WireFrame) != 0); }
            set { if (Wireframe != value) RenderFlags ^= RenderFlags.WireFrame; }
        }

        public bool Scaling
        {
            get { return ((RenderFlags & RenderFlags.Scaling) != 0); }
            set { if (Scaling != value) RenderFlags ^= RenderFlags.Scaling; }
        }

        #endregion

        #region Control Modes

        public void ResetControlStates()
        {
            SetupControlStates();
        }

        #endregion

        #region Event Handlers

        void ModelExView_ChangeFocus(object sender, EventArgs e)
        {
        }

        void ModelExView_PreviewKeyDown(object sender, PreviewKeyDownEventArgs e)
        {
            if (ArcBall != null && Camera != null)
            {
                #region Dragging not allowed

                if (ControlMode == ControlType.OrbitCamera ||
                    ControlMode == ControlType.FlyCamera)
                {

                    if (!ArcBall.IsMoving)
                    {
                        switch (e.KeyCode)
                        {
                            case Keys.Up:
                                Camera.RotateY(2.0f * ((float)Math.PI) / -200.0f);
                                break;
                            case Keys.Down:
                                Camera.RotateY(2.0f * ((float)Math.PI) / 200.0f);
                                break;
                            case Keys.Left:
                            {
                                if (e.Control)
                                {
                                    Camera.RotateZ(2.0f * ((float)Math.PI) / 200.0f);
                                }
                                else
                                {
                                    Camera.RotateX(2.0f * ((float)Math.PI) / -200.0f);
                                }
                                break;
                            }
                            case Keys.Right:
                            {
                                if (e.Control)
                                {
                                    Camera.RotateZ(2.0f * ((float)Math.PI) / -200.0f);
                                }
                                else
                                {
                                    Camera.RotateX(2.0f * ((float)Math.PI) / 200.0f);
                                }
                                break;
                            }
                        }
                    }
                }

                #endregion

                #region Dragging allowed

                if (ControlMode == ControlType.OrbitCamera ||
                    ControlMode == ControlType.FlyCamera ||
                    ControlMode == ControlType.MapCamera)
                {
                    switch (e.KeyCode)
                    {
                        case Keys.W:
                        {
                            if (Camera.Mode == Camera.CameraMode.Fly)
                            {
                                Camera.MoveForward(0.01f);
                            }
                            else if (Camera.Mode == Camera.CameraMode.Map)
                            {
                                Camera.MoveUpward(0.01f);
                            }
                            break;
                        }
                        case Keys.S:
                        {
                            if (Camera.Mode == Camera.CameraMode.Fly)
                            {
                                Camera.MoveForward(-0.01f);
                            }
                            else if (Camera.Mode == Camera.CameraMode.Map)
                            {
                                Camera.MoveUpward(-0.01f);
                            }
                            break;
                        }
                        case Keys.A:
                        {
                            if (Camera.Mode == Camera.CameraMode.Fly ||
                                Camera.Mode == Camera.CameraMode.Map)
                            {
                                Camera.MoveRight(-0.01f);
                            }
                            break;
                        }
                        case Keys.D:
                        {
                            if (Camera.Mode == Camera.CameraMode.Fly ||
                                Camera.Mode == Camera.CameraMode.Map)
                            {
                                Camera.MoveRight(0.01f);
                            }
                            break;
                        }
                        case Keys.U:
                        {
                            if (Camera.Mode == Camera.CameraMode.Fly ||
                                Camera.Mode == Camera.CameraMode.Map)
                            {
                                Camera.RemoveRoll();
                            }
                            break;
                        }
                    }
                }

                #endregion
            }
        }

        void ModelExView_MouseWheel(object sender, MouseEventArgs e)
        {
        }

        void ModelExView_MouseUp(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (ArcBall != null)
                {
                    ArcBall.EndMove();
                }

                Menu.HandleMouseClick(LastMousePosition);
                ScrollBar.HandleMouseUp(LastMousePosition);
            }
        }

        void ModelExView_MouseDown(object sender, MouseEventArgs e)
        {
            if (e.Button == MouseButtons.Left)
            {
                if (ArcBall != null)
                {
                    ArcBall.BeginMove();
                }
                LastMouseDownPosition = new Point(e.X, e.Y);
                ScrollBar.HandleMouseDown(LastMousePosition);
            }
        }

        void ModelExView_MouseMove(object sender, MouseEventArgs e)
        {
            Point xLocation = new Point(e.X, e.Y);

            if (e.Button == MouseButtons.Left)
            {
                if (xLocation.X < Left) xLocation.X = Left;
                if (xLocation.X > Right) xLocation.X = Right;
                if (xLocation.Y < Top) xLocation.Y = Top;
                if (xLocation.Y > Bottom) xLocation.Y = Bottom;

                if (ControlMode == ControlType.MapCamera)
                {
                    float fAspectRatio = (float)Width / (float)Height;
                    Vector3 xDirection = Camera.Target - Camera.Position;
                    float fDistance = Math.Abs(xDirection.Length());
                    float fHeight = 2.0f * fDistance * (float)Math.Tan(Math.PI / 8.0f);
                    float fPixelSize = fHeight / Height;

                    if (Camera != null)
                    {
                        Camera.MoveRight((xLocation.X - LastMousePosition.X) * -fPixelSize);
                        Camera.MoveUpward((xLocation.Y - LastMousePosition.Y) * fPixelSize);
                    }
                }
                else
                {
                    if (ArcBall != null)
                    {
                        ArcBall.OnMove(
                            xLocation.X - Left,
                            xLocation.Y - Top,
                            (ControlMode == ControlType.RotateObject) ? Camera.Rotation : Quaternion.Identity
                        );
                    }
                }
            }

            LastMousePosition = xLocation;
            Menu.HandleMouseMove(LastMousePosition);
            ScrollBar.HandleMouseMove(LastMousePosition);
        }

        bool ModelExView_Open(ModelExListBox listBox)
        {
            PauseRendering = true;
            OpenFileDialog OpenDlg = new OpenFileDialog();
            OpenDlg.CheckFileExists = true;
            OpenDlg.CheckPathExists = true;
            OpenDlg.Filter =    "Soul Reaver Mesh Files|*.SRObj;*.drm;*.pcm|" +
                                "Soul Reaver DRM Files (*.drm)|*.drm|" +
                                "Soul Reaver PCM Files (*.pcm)|*.pcm|" +
                                "Collada Mesh Files (*.dae)|*.dae|" +
                                "All Mesh Files|*.SRObj;*.drm;*.pcm|" +
                                "All Files (*.*)|*.*";
            OpenDlg.DefaultExt = "drm";
            OpenDlg.FilterIndex = 1;
            if (OpenDlg.ShowDialog() == DialogResult.OK)
            {
                if (Path.GetExtension(OpenDlg.FileName) == ".dae")
                {
                    Project.LoadColladaFile(OpenDlg.FileName);
                }
                else
                {
                    Control xControl = this;
                    Project.LoadSoulReaverFile(OpenDlg.FileName);
                    SelectedObject = Project.FindObject(Project.Name);
                    SelectedObject.ArcBall = new ArcBall(ref xControl, false);
                }
            }

            PauseRendering = false;
            return true;
        }

        bool ModelExView_Save(ModelExListBox listBox)
        {
            PauseRendering = true;
            SaveFileDialog SaveDlg = new SaveFileDialog();
            SaveDlg.CheckFileExists = false;
            SaveDlg.CheckPathExists = false;
            SaveDlg.Filter = "Collada Document (*.dae)|*.dae";
            SaveDlg.DefaultExt = "dae";
            SaveDlg.FilterIndex = 1;
            SaveDlg.FileName = Project.Name + ".dae";
            if (SaveDlg.ShowDialog() == DialogResult.OK)
            {
                Project.SaveColladaFile(SaveDlg.FileName);
            }

            PauseRendering = false;
            return true;
        }

        bool ModelExView_FromFile(ModelExListBox listBox)
        {
            PauseRendering = true;
            OpenFileDialog OpenDlg = new OpenFileDialog();
            if (OpenDlg.ShowDialog() == DialogResult.OK)
            {
            }
            PauseRendering = false;
            return true;
        }

        bool ModelExView_FromFolder(ModelExListBox listBox)
        {
            PauseRendering = true;
            FolderBrowserDialog xFolderDlg = new FolderBrowserDialog();
            xFolderDlg.Description = "Load all required textures from the current folder";
            xFolderDlg.ShowNewFolderButton = false;
            if (xFolderDlg.ShowDialog() == DialogResult.OK)
            {
            }
            PauseRendering = false;
            return true;
        }

        bool ModelExView_Exit(ModelExListBox listBox)
        {
            ((Form)TopLevelControl).Close();
            return true;
        }

        bool ModelExView_ToggleLighting(ModelExListBox listBox)
        {
            Lighting ^= true;
            listBox.SelectedItem.StatusBox.Status = (Lighting) ? 1 : 0;
            return true;
        }

        bool ModelExView_ToggleTextures(ModelExListBox listBox)
        {
            Textures ^= true;
            listBox.SelectedItem.StatusBox.Status = (Textures) ? 1 : 0;
            return true;
        }

        bool ModelExView_ToggleTextureFiltering(ModelExListBox listBox)
        {
            TextureFiltering ^= true;
            listBox.SelectedItem.StatusBox.Status = (TextureFiltering) ? 1 : 0;
            return true;
        }

        bool ModelExView_ToggleWireframe(ModelExListBox listBox)
        {
            Wireframe ^= true;
            listBox.SelectedItem.StatusBox.Status = (Wireframe) ? 1 : 0;
            return true;
        }

        bool ModelExView_SelectDebugOutput(ModelExListBox listBox)
        {
            for (int i = 0; i < listBox.Items.Count; i++)
            {
                if (listBox.Items[i] == listBox.SelectedItem)
                {
                    listBox.Items[i].StatusBox.Status = 1;
                    switch (i)
                    {
                        case 0:
                            DebugMode = DebugType.Debug;
                            break;
                        case 1:
                            DebugMode = DebugType.Textures;
                            break;
                        case 2:
                            DebugMode = DebugType.Effects;
                            break;
                        case 3:
                            DebugMode = DebugType.Materials;
                            break;
                        case 4:
                            DebugMode = DebugType.None;
                            break;
                    }
                }
                else
                {
                    listBox.Items[i].StatusBox.Status = 0;
                }
            }

            return true;
        }

        #endregion

        protected override void OnTick()
        {
            base.OnTick();
        }

        protected override void RenderScene()
        {
            base.RenderScene();

            m_xDevice.BeginScene();

            if (Menu != null)
            {
                UpdateMenus();
                Menu.Render();
            }

            if (DebugOutput != null)
            {
                if (DebugMode == DebugType.Debug)
                {
                    UpdateDebugOutput();
                    DebugOutput.Render();
                }
            }

            if (TextureOutput != null)
            {
                if (DebugMode == DebugType.Textures)
                {
                    UpdateTextureOutput();
                    TextureOutput.Render();
                }
            }

            if (MaterialOutput != null)
            {
                if (DebugMode == DebugType.Materials)
                {
                    UpdateMaterialOutput();
                    MaterialOutput.Render();

                    if (ScrollBar != null)
                    {
                        ScrollBar.Render();
                    }
                }
            }

            if (EffectOutput != null)
            {
                if (DebugMode == DebugType.Effects)
                {
                    UpdateEffectOutput();
                    EffectOutput.Render();
                }
            }

            m_xDevice.EndScene();

            if (Camera != null)
            {
                Camera.Update();
            }

            UpdateControlStates();

            m_xDevice.BeginScene();

            //float scale = 0.335799217f;
            //if (m_eControlMode == ControlType.MapCamera)
            //{
            //    scale = 0.06109793f;
            //}

            float scale = 0.05f;
            Project.RenderObjects(RenderFlags, CullMode, scale);

            m_xDevice.EndScene();
        }

        protected void UpdateMenus()
        {
            if (Menu == null)
            {
                return;
            }

            Menu.Rect.Width = Width;
        }

        protected void UpdateDebugOutput()
        {
            if (DebugOutput == null)
            {
                return;
            }

            DebugOutput.Rect.X = 0;
            DebugOutput.Rect.Y = Height - DebugOutput.Rect.Height;
            DebugOutput.Rect.Width = Width;
            DebugOutput.UpdatePositions();

            DebugOutput.Text = "";

            DebugOutput.Text += "App has focus = ";
            DebugOutput.Text += (TopLevelControl != null && TopLevelControl.ContainsFocus == true) ? "true" : "false";
            DebugOutput.Text += "\r\n";

            DebugOutput.Text += "Render view has focus = ";
            DebugOutput.Text += (Focused == true) ? "true" : "false";
            DebugOutput.Text += "\r\n";

            DebugOutput.Text += "Last device reset: " + LastDeviceReset;
            DebugOutput.Text += "\r\n";

            DebugOutput.Text += "Last mouse position: " + LastMousePosition.ToString();
            DebugOutput.Text += "\r\n";
            DebugOutput.Text += "Last click position: " + LastMouseDownPosition.ToString();
            DebugOutput.Text += "\r\n";
            DebugOutput.Text += "Frame rate: " + FrameRate + " fps";
            DebugOutput.Text += "\r\n";
            DebugOutput.Text += "Time of last frame: " + LastFrameTime;
            DebugOutput.Text += "\r\n";

            if (ArcBall != null)
            {
                DebugOutput.Text += "Dragging arcball = ";
                DebugOutput.Text += (ArcBall.IsMoving == true) ? "true" : "false";
                DebugOutput.Text += "\r\n";
            }

            if (Camera != null)
            {
                DebugOutput.Text += String.Format(
                    "Camera Rotation: {{ W = {0:F2}, X = {0:F2}, Y = {1:F2}, Z = {1:F2} }}\r\n",
                    Camera.Rotation.W,
                    Camera.Rotation.X,
                    Camera.Rotation.Y,
                    Camera.Rotation.Z
                );
            }
        }

        protected void UpdateTextureOutput()
        {
            if (TextureOutput == null)
            {
                return;
            }

            TextureOutput.Rect.X = 0;
            TextureOutput.Rect.Y = Height - TextureOutput.Rect.Height;
            TextureOutput.Rect.Width = Width;
            TextureOutput.UpdatePositions();

            TextureOutput.Text = Project.ListTextures();
        }

        protected void UpdateMaterialOutput()
        {
            if (MaterialOutput == null)
            {
                return;
            }

            MaterialOutput.Rect.X = 0;
            MaterialOutput.Rect.Y = Height - MaterialOutput.Rect.Height;
            MaterialOutput.Rect.Width = Width;
            MaterialOutput.UpdatePositions();

            ScrollBar.Rect.X = MaterialOutput.Rect.Right - ScrollBar.Rect.Width;
            ScrollBar.Rect.Y = MaterialOutput.Rect.Y;
            ScrollBar.UpdatePositions();
            MaterialOutput.Offset.Y = -ScrollBar.Position;

            MaterialOutput.Text = Project.ListMaterials();
        }

        protected void UpdateEffectOutput()
        {
            if (EffectOutput == null)
            {
                return;
            }

            EffectOutput.Rect.X = 0;
            EffectOutput.Rect.Y = Height - EffectOutput.Rect.Height;
            EffectOutput.Rect.Width = Width;
            EffectOutput.UpdatePositions();

            EffectOutput.Text = Project.ListEffects();
        }
    }
}