﻿using System;
using System.Collections.Generic;
using System.Drawing;
using Microsoft.DirectX;
using Microsoft.DirectX.Direct3D;
using DX = Microsoft.DirectX;
using D3DX = Microsoft.DirectX.Direct3D;

namespace ModelEx
{
    class ModelExListBox : ModelExTextBox
    {
        public delegate bool ItemClickHandler(ModelExListBox listBox);

        public class Item
        {
            public ModelExStatusBox StatusBox;
            public ModelExListBox ListBox;
            public ItemClickHandler ClickHandler;
            protected bool _Enabled;

            public bool Enabled
            {
                get
                {
                    return _Enabled;
                }
                set
                {
                    _Enabled = value;
                    if (StatusBox != null)
                    {
                        StatusBox.Enabled = _Enabled;
                    }
                    if (ListBox != null)
                    {
                        ListBox.Enabled = _Enabled;
                    }
                }
            }
        }

        ModelExFont StatusFont;
        public List<Item> Items;
        public Item SelectedItem;
        Int32 ItemWidth;
        Int32 ItemHeight;
        bool Vertical;
        bool AutoExpand;
        bool Expanded;

        public ModelExListBox(Device device, Rectangle rect, UInt32 colour, UInt32 borderColour, UInt32 textColour, UInt32 textColourDisabled, ModelExFont font, ModelExFont statusFont, Point margin, Int32 itemWidth, Int32 itemHeight, bool vertical) :
            base(device, rect, colour, borderColour, textColour, textColourDisabled, font, margin)
        {
            StatusFont = statusFont;
            Items = new List<Item>();
            ItemWidth = itemWidth;
            ItemHeight = itemHeight;
            Vertical = vertical;
            AutoExpand = vertical;
        }

        public void AddItem(String name, UInt32 index, ModelExListBox listBox, String[] statusText, int status)
        {
            Item newEntry = new Item();
            newEntry.StatusBox = new ModelExStatusBox(Device, name, new Rectangle(), Colour, BorderColour, TextColour, TextColourDisabled, Font, Margin, StatusFont, Margin, statusText);
            newEntry.StatusBox.Status = status;
            newEntry.ListBox = listBox;
            newEntry.Enabled = true;
            Items.Insert((Int32)index, newEntry);
            UpdatePositions();
        }

        public void AddItem(String name, UInt32 index, ModelExListBox listBox)
        {
            Item newEntry = new Item();
            newEntry.StatusBox = new ModelExStatusBox(Device, name, new Rectangle(), Colour, BorderColour, TextColour, TextColourDisabled, Font, Margin, StatusFont, Margin, null);
            newEntry.ListBox = listBox;
            newEntry.Enabled = true;
            Items.Insert((Int32)index, newEntry);
            UpdatePositions();
        }

        public void AddItem(String name, UInt32 index, ModelExListBox listBox, String[] statusText, int status, ItemClickHandler clickHandler)
        {
            Item newEntry = new Item();
            newEntry.StatusBox = new ModelExStatusBox(Device, name, new Rectangle(), Colour, BorderColour, TextColour, TextColourDisabled, Font, Margin, StatusFont, Margin, statusText);
            newEntry.StatusBox.Status = status;
            newEntry.ListBox = listBox;
            newEntry.Enabled = true;
            newEntry.ClickHandler = clickHandler;
            Items.Insert((Int32)index, newEntry);
            UpdatePositions();
        }

        public void AddItem(String name, UInt32 index, ModelExListBox listBox, ItemClickHandler clickHandler)
        {
            Item newEntry = new Item();
            newEntry.StatusBox = new ModelExStatusBox(Device, name, new Rectangle(), Colour, BorderColour, TextColour, TextColourDisabled, Font, Margin, StatusFont, Margin, null);
            newEntry.ListBox = listBox;
            newEntry.ClickHandler = clickHandler;
            newEntry.Enabled = true;
            Items.Insert((Int32)index, newEntry);
            UpdatePositions();
        }

        public void RemoveItem(UInt32 index)
        {
            Items.RemoveAt((Int32)index);
            UpdatePositions();
        }

        public new void UpdatePositions()
        {
            base.UpdatePositions();

            if (Vertical)
            {
                Int32 itemWidth = 10;
                bool hasStatusText = false;
                foreach (Item item in Items)
                {
                    itemWidth = Math.Max(Font.MeasureString(item.StatusBox.Text).Width, itemWidth);
                    if (item.StatusBox.StatusText != null)
                    {
                        hasStatusText = true;
                    }
                }

                itemWidth += Margin.X * 2;
                if (hasStatusText)
                {
                    itemWidth += ItemHeight;
                }

                ItemWidth = itemWidth;
                Rect.Width = itemWidth;
            }

            Item prevItem = null;
            foreach (Item item in Items)
            {
                Int32 itemWidth = ItemWidth;
                if (!Vertical)
                {
                    itemWidth = Font.MeasureString(item.StatusBox.Text).Width;
                    itemWidth += Margin.X * 2;
                    if (item.StatusBox.StatusText != null)
                    {
                        itemWidth += ItemHeight;
                    }
                }

                Rectangle itemRect = new Rectangle(Rect.Left, Rect.Top, itemWidth, ItemHeight);
                if (prevItem != null)
                {
                    if (Vertical)
                    {
                        itemRect.Location = new Point(Rect.Left, prevItem.StatusBox.Rect.Bottom);
                    }
                    else
                    {
                        itemRect.Location = new Point(prevItem.StatusBox.Rect.Right, Rect.Top);
                    }
                }

                item.StatusBox.Rect = itemRect;
                item.StatusBox.UpdatePositions();
                prevItem = item;
            }

            if (SelectedItem != null && SelectedItem.ListBox != null)
            {
                if (Vertical)
                {
                    SelectedItem.ListBox.Rect.Location = new Point(SelectedItem.StatusBox.Rect.Right, SelectedItem.StatusBox.Rect.Top);
                }
                else
                {
                    SelectedItem.ListBox.Rect.Location = new Point(SelectedItem.StatusBox.Rect.Left, SelectedItem.StatusBox.Rect.Bottom);
                }

                SelectedItem.ListBox.UpdatePositions();
            }
        }

        public bool HandleMouseMove(Point mousePosition)
        {
            bool isOpen = false;
            foreach (Item item in Items)
            {
                item.StatusBox.Colour = Colour;
                item.StatusBox.TextColour = TextColour;

                if (item.StatusBox.Rect.Contains(mousePosition))
                {
                    item.StatusBox.Colour ^= 0x00FFFFFF;
                    item.StatusBox.TextColour ^= 0x00FFFFFF;

                    if (SelectedItem != item)
                    {
                        SelectedItem = item;
                        Expanded |= AutoExpand;
                    }

                    isOpen = true;
                }
            }

            if (Expanded &&
                SelectedItem != null &&
                SelectedItem.ListBox != null &&
                SelectedItem.ListBox.HandleMouseMove(mousePosition))
            {
                isOpen = true;
            }

            UpdatePositions();

            return isOpen;
        }

        public bool HandleMouseClick(Point mousePosition)
        {
            bool wasClicked = false;

            if (Enabled)
            {
                if (SelectedItem != null)
                {
                    if (Expanded &&
                        (SelectedItem.Enabled) && 
                        SelectedItem.ListBox != null &&
                        SelectedItem.ListBox.HandleMouseClick(mousePosition))
                    {
                        wasClicked = true;
                    }
                    else if (SelectedItem.StatusBox.Rect.Contains(mousePosition))
                    {
                        Expanded = true;
                        wasClicked = true;
                    }
                }

                if (!wasClicked)
                {
                    Expanded = false;
                    SelectedItem = null;
                }
                else if ((SelectedItem.ClickHandler != null) && (SelectedItem.Enabled))
                {
                    SelectedItem.ClickHandler(this);
                    Expanded = false;
                    SelectedItem = null;
                    wasClicked = false;
                }
            }

            return wasClicked;
        }

        public override void Render()
        {
            base.Render();

            foreach (Item item in Items)
            {
                item.StatusBox.Render();

                if (Expanded &&
                    SelectedItem != null &&
                    SelectedItem.ListBox != null)
                {
                    SelectedItem.ListBox.Render();
                }
            }
        }
    }
}
